package net.knarcraft.blacksmithvisuals.listener;

import com.comphenix.protocol.PacketType;
import com.comphenix.protocol.ProtocolLibrary;
import com.comphenix.protocol.ProtocolManager;
import com.comphenix.protocol.events.PacketContainer;
import net.citizensnpcs.api.npc.NPC;
import net.knarcraft.blacksmith.event.ActionStartEvent;
import net.knarcraft.blacksmith.event.BlacksmithReforgeFailEvent;
import net.knarcraft.blacksmith.event.BlacksmithReforgeStartEvent;
import net.knarcraft.blacksmith.event.BlacksmithReforgeSucceedEvent;
import net.knarcraft.blacksmith.event.NPCSoundEvent;
import net.knarcraft.blacksmith.event.ScrapperSalvageFailEvent;
import net.knarcraft.blacksmith.event.ScrapperSalvageStartEvent;
import net.knarcraft.blacksmith.event.ScrapperSalvageSucceedEvent;
import net.knarcraft.blacksmithvisuals.BlacksmithVisuals;
import net.knarcraft.blacksmithvisuals.container.AnimationData;
import net.knarcraft.blacksmithvisuals.container.NPCData;
import net.knarcraft.blacksmithvisuals.container.SoundData;
import net.knarcraft.blacksmithvisuals.manager.ConfigurationManager;
import net.knarcraft.blacksmithvisuals.property.NPCPosition;
import net.knarcraft.blacksmithvisuals.property.SoundIdentifier;
import org.bukkit.Bukkit;
import org.bukkit.Location;
import org.bukkit.World;
import org.bukkit.entity.Entity;
import org.bukkit.entity.Player;
import org.bukkit.event.EventHandler;
import org.bukkit.event.Listener;
import org.bukkit.scheduler.BukkitScheduler;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.Random;
import java.util.logging.Level;

/**
 * A listener for blacksmith-related events
 */
public class BlacksmithListener implements Listener {

    private final Random random;
    private final ConfigurationManager configurationManager;

    /**
     * Instantiates a new blacksmith listener
     *
     * @param configurationManager <p>The configuration manager to get configuration values from</p>
     */
    public BlacksmithListener(@NotNull ConfigurationManager configurationManager) {
        this.random = new Random();
        this.configurationManager = configurationManager;
    }

    @EventHandler
    public void onDefaultSound(@NotNull NPCSoundEvent event) {
        event.setCancelled(true);
    }

    @EventHandler
    public void onReforgeStart(@NotNull BlacksmithReforgeStartEvent event) {
        runWorkingAnimation(event, SoundIdentifier.REFORGING_WORKING, this.configurationManager.getBlacksmithAnimationData());
    }

    @EventHandler
    public void onSalvageStart(@NotNull ScrapperSalvageStartEvent event) {
        runWorkingAnimation(event, SoundIdentifier.SALVAGING_WORKING, this.configurationManager.getScrapperAnimationData());
    }

    @EventHandler
    public void onReforgeSuccess(@NotNull BlacksmithReforgeSucceedEvent event) {
        playSound(event.getNpc().getEntity(), this.configurationManager.getSoundData(SoundIdentifier.REFORGING_SUCCESS));
    }

    @EventHandler
    public void onSalvageSuccess(@NotNull ScrapperSalvageSucceedEvent event) {
        playSound(event.getNpc().getEntity(), this.configurationManager.getSoundData(SoundIdentifier.SALVAGING_SUCCESS));
    }

    @EventHandler
    public void onReforgeFail(@NotNull BlacksmithReforgeFailEvent event) {
        playSound(event.getNpc().getEntity(), this.configurationManager.getSoundData(SoundIdentifier.REFORGING_FAILURE));
    }

    @EventHandler
    public void onSalvageFail(@NotNull ScrapperSalvageFailEvent event) {
        playSound(event.getNpc().getEntity(), this.configurationManager.getSoundData(SoundIdentifier.SALVAGING_FAILURE));
    }

    /**
     * Runs the working animation for an NPC
     *
     * @param event           <p>The action that started</p>
     * @param soundIdentifier <p>The identifier for the sound to play</p>
     * @param animationData   <p>The animation data for the animation to play</p>
     */
    private void runWorkingAnimation(@NotNull ActionStartEvent event, @NotNull SoundIdentifier soundIdentifier,
                                     @NotNull AnimationData animationData) {
        BlacksmithVisuals instance = BlacksmithVisuals.getInstance();
        BukkitScheduler scheduler = Bukkit.getScheduler();
        NPC npc = event.getNpc();
        long delay = moveToWorkingPosition(npc, NPCPosition.getFromMaterial(event.getCraftingStation()));
        long finishTime = event.getActionDurationTicks() - (2 * delay);

        scheduler.scheduleSyncDelayedTask(instance, () -> startWorkAnimation(npc.getEntity(),
                this.configurationManager.getSoundData(soundIdentifier), animationData, finishTime - 20), delay);
        scheduler.scheduleSyncDelayedTask(instance, () -> moveBack(event.getNpc()), finishTime);
    }

    /**
     * Moves an NPC back to its idle position
     *
     * @param npc <p>The NPC to move</p>
     */
    private void moveBack(@NotNull NPC npc) {
        if (!npc.isSpawned()) {
            return;
        }

        NPCData npcData = BlacksmithVisuals.getInstance().getNpcDataManager().getData(npc.getUniqueId());
        if (npcData == null) {
            return;
        }

        Location targetLocation = npcData.positions().get(NPCPosition.IDLE);
        if (!npc.getNavigator().canNavigateTo(targetLocation)) {
            BlacksmithVisuals.getInstance().getLogger().log(Level.WARNING, "Idle position for NPC " +
                    npc.getName() + " is unreachable!");
            return;
        }
        npc.getNavigator().setTarget(targetLocation);

        Bukkit.getScheduler().scheduleSyncDelayedTask(BlacksmithVisuals.getInstance(), () ->
                npc.getEntity().teleport(targetLocation), getWalkTime(npc.getEntity().getLocation(), targetLocation));
    }

    /**
     * Moves a npc to its working position
     *
     * @param npc         <p>The NPC to move</p>
     * @param npcPosition <p>The npc position to move to</p>
     * @return <p>The time the move will take</p>
     */
    private long moveToWorkingPosition(@NotNull NPC npc, @Nullable NPCPosition npcPosition) {
        if (!npc.isSpawned() || npcPosition == null) {
            return 0;
        }

        NPCData npcData = BlacksmithVisuals.getInstance().getNpcDataManager().getData(npc.getUniqueId());
        if (npcData == null) {
            return 0;
        }

        Location targetLocation = npcData.positions().get(npcPosition);
        if (targetLocation == null && npcPosition != NPCPosition.WORKING_REPAIRABLE) {
            targetLocation = npcData.positions().get(NPCPosition.WORKING_REPAIRABLE);
        }
        if (targetLocation == null) {
            return 0;
        }

        if (!npc.getNavigator().canNavigateTo(targetLocation)) {
            BlacksmithVisuals.getInstance().getLogger().log(Level.WARNING, "Working position for NPC " +
                    npc.getName() + " is unreachable!");
            return 0;
        }

        // Move NPC using Citizens path-finding
        npc.getNavigator().setTarget(targetLocation);

        // Teleport the NPC tp get it in the exact final location
        long walkTime = getWalkTime(npc.getEntity().getLocation(), targetLocation);
        Location finalTargetLocation = targetLocation;
        Bukkit.getScheduler().scheduleSyncDelayedTask(BlacksmithVisuals.getInstance(), () ->
                npc.getEntity().teleport(finalTargetLocation), walkTime);
        return walkTime;
    }

    /**
     * Starts the working animation and sound
     *
     * @param entity        <p>The entity to play the sound at</p>
     * @param soundData     <p>The sound data for the sound to play</p>
     * @param animationData <p>The animation data for the animation to play</p>
     * @param durationTicks <p>The duration of the work animation</p>
     */
    private void startWorkAnimation(@NotNull Entity entity, @NotNull SoundData soundData,
                                    @NotNull AnimationData animationData, long durationTicks) {
        BlacksmithVisuals instance = BlacksmithVisuals.getInstance();
        BukkitScheduler scheduler = Bukkit.getScheduler();

        int playWorkSound = scheduler.scheduleSyncRepeatingTask(instance,
                () -> animateNPC(entity, soundData, animationData), 20, animationData.animationDelay());
        scheduler.scheduleSyncDelayedTask(instance, () -> scheduler.cancelTask(playWorkSound), durationTicks);
    }

    /**
     * Gets the time it would take to go from one location to another in ticks
     *
     * @param startLocation  <p>The location to start from</p>
     * @param targetLocation <p>The target location</p>
     * @return <p>The time in ticks</p>
     */
    private long getWalkTime(@NotNull Location startLocation, @NotNull Location targetLocation) {
        double distance = startLocation.distance(targetLocation);
        double WALK_SPEED = 4.317;
        return Math.round((distance / WALK_SPEED) * 20);
    }

    /**
     * Animates a npc's hand, and plays a sound
     *
     * @param entity        <p>The entity to animate and play the sound at</p>
     * @param soundData     <p>The sound to be played</p>
     * @param animationData <p>The animation to be played</p>
     */
    private void animateNPC(@NotNull Entity entity, @NotNull SoundData soundData, @NotNull AnimationData animationData) {
        if (random.nextInt(100) >= animationData.animationChance()) {
            return;
        }

        playSound(entity, soundData);

        // Don't play disabled animations
        if (!animationData.animateOffHand()) {
            return;
        }

        if (soundData.offsetTicks() < 0) {
            Bukkit.getScheduler().scheduleSyncDelayedTask(BlacksmithVisuals.getInstance(),
                    () -> animateOffhand(entity), -soundData.offsetTicks());
        } else {
            animateOffhand(entity);
        }
    }

    /**
     * Animates an NPC's offhand
     *
     * @param entity <p>The entity to animate</p>
     */
    private void animateOffhand(@NotNull Entity entity) {
        ProtocolManager manager = ProtocolLibrary.getProtocolManager();
        PacketContainer packet = manager.createPacket(PacketType.Play.Server.ANIMATION);
        packet.getIntegers().write(0, entity.getEntityId());
        packet.getIntegers().write(1, 3);

        for (Player player : Bukkit.getOnlinePlayers()) {
            if (player.getWorld().equals(entity.getWorld())) {
                manager.sendServerPacket(player, packet);
            }
        }
    }

    /**
     * Plays a sound according to the given sound data
     *
     * @param entity    <p>The entity to play the sound from</p>
     * @param soundData <p>The data describing the sound to play</p>
     */
    private void playSound(@NotNull Entity entity, @NotNull SoundData soundData) {
        // Don't play disabled sounds
        if (!soundData.enabled()) {
            return;
        }

        World world = entity.getLocation().getWorld();
        if (world == null) {
            return;
        }

        int delay = Math.max(soundData.offsetTicks(), 0);
        Bukkit.getScheduler().scheduleSyncDelayedTask(BlacksmithVisuals.getInstance(), () ->
                world.playSound(entity, soundData.sound(), soundData.soundCategory(),
                        soundData.volume(), soundData.pitch()), delay);
    }

}
